#!/usr/bin/env python3
# -*- coding: utf-8 -*-

category = 'Spectrometer software'

class Driver_parser():
    def __init__(self, Instance, name, **kwargs):
        self.name     = name
        self.Instance = Instance
        
        
    def add_parser_usage(self,message):
        """Usage to be used by the parser"""
        usage = f"""
{message}

----------------  Examples:  ----------------

usage:    autolab driver [options] args
            
    autolab driver -D {self.name} -C SOCKET -A 192.168.0.2 -e 0.1 -o my_file_name
    load {self.name} driver using SOCKET communication protocol with address 192.168.0.2 and acquire data after 2 expositions of 0.1s each. Save those into my_file_name.txt.
    
    autolab driver -D nickname -a -o my_file_name
    Similar to previous one using device's nickname defined in local_config.ini, turn on auto vertical scale/exposure optimization loop (so as to maximize signal vertically), get and save the data.
            """
        return usage
    
    def add_parser_arguments(self,parser):
        """Add arguments to the parser passed as input"""
        parser.add_argument("-e", "--exposure", type=str, dest="exposure", default=None, help="Set the exposure time in s")
        parser.add_argument("-o", "--filename", type=str, dest="filename", default=None, help="Set the name of the output file")
        parser.add_argument("-a", "--auto_exposure", action = "store_true", dest="auto_exposure", default=False, help="Get data from camera using auto exposition mode (maximize signal vertically by adjusting the exposure)")
        parser.add_argument("-F", "--force",action = "store_true", dest="force", default=False, help="Allows overwriting file")
        
        return parser

    def do_something(self,args):
        if args.exposure:
            getattr(self.Instance,'setExposureTime')(args.exposure)
        if args.auto_exposure:
            getattr(self.Instance,'setAutoExposureTimeEnabled')(True)
        if args.filename:
            getattr(self.Instance,'acquireSpectrum')()
            data = getattr(self.Instance,'data')['spectrum']
            temp_filename = f'{filename}_WINSPEC.txt'
            if os.path.exists(os.path.join(os.getcwd(),temp_filename)) and not(args.force):
                print(f'\nFile {temp_filename} already exists, change filename or remove old file\n')
                return
            f = savetxt(data)
            print('WINSPEC data saved')
        

    def exit(self):
        self.Instance.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

category = 'Optical source'
    

class Driver_parser():
    def __init__(self, Instance, name, **kwargs):
        self.name     = name
        self.Instance = Instance
        
        
    def add_parser_usage(self,message):
        """Usage to be used by the parser"""
        usage = f"""
{message}

----------------  Examples:  ----------------

usage:    autolab driver [options] args
            
    autolab driver -D {self.name} -A GPIB::7::INSTR -C VISA -a 30 -t 20.1
    load {self.name} driver using VISA communication protocol with address TCPIP... and set the current to 30mA and the temperature to 20.1 degree celcius.
    
    autolab driver -D nickname -p 30 -t 20.1
    same as before but using the device nickname as defined in local_config.ini and set the power setpoint to 30mW
            """
        return usage
    
    def add_parser_arguments(self,parser):
        """Add arguments to the parser passed as input"""
        parser.add_argument("-a", "--current", type=str, dest="current", default=None, help="Set the current in mA." )
        parser.add_argument("-p", "--power", type=str, dest="power", default=None, help="Set the pump power setpoint in mW." )
        parser.add_argument("-t", "--temperature", type=str, dest="temperature", default=None, help="Set locking temperature." )
        
        return parser

    def do_something(self,args):
        if args.current:
            getattr(getattr(self.Instance,'las'),'set_current')(args.current)
        elif args.power:
            getattr(getattr(self.Instance,'las'),'set_power')(args.power)
        if args.temperature:
            getattr(getattr(self.Instance,'tec'),'set_temperature')(args.temperature)
            

    def exit(self):
        self.Instance.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

category = 'Electrical source'
    
    
class Driver_parser():
    def __init__(self, Instance, name, **kwargs):
        self.name     = name
        self.Instance = Instance
        
        
    def add_parser_usage(self,message):
        """Usage to be used by the parser"""
        usage = f"""
{message}

----------------  Examples:  ----------------

usage:    autolab driver [options] args
            
    autolab driver -D {self.name} -A GPIB::7::INSTR -C VISA -v 0.2
    load {self.name} driver using VISA communication protocol with address GPIB... and set the voltage to 0.2V 
    
    autolab driver -D nickname -a 0.5
    using the device nickname as defined in local_config.ini modify the current compliance to 0.5 A
    
    autolab driver -D nickname -s 0
    using the device nickname as defined in local_config.ini, turn the output OFF
    
            """
        return usage
    
    def add_parser_arguments(self,parser):
        """Add arguments to the parser passed as input"""
        parser.add_argument("-v", "--voltage", type=str, dest="voltage", default=None, help="Set the voltage in V." )
        parser.add_argument("-a", "--current", type=str, dest="current", default=None, help="Set the current in A." )
        parser.add_argument("-s", "--state", type=str, dest="state", default=None, help="Turn output OFF or ON: arguments are 0 or 1 (for OFF and ON, respectively)." )
        
        return parser

    def do_something(self,args):
        if args.voltage:
            getattr(self.Instance,'set_voltage_compliance')(args.voltage)
        if args.current:
            getattr(self.Instance,'set_current_compliance')(args.current)
        if args.state:
            assert isinstance(bool(int(float(args.state))),bool)
            args.state = bool(int(float(args.state)))
            getattr(self.Instance,'set_output_state')(args.state)
            
    def exit(self):
        self.Instance.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- 
"""

import os
from numpy import savetxt,linspace
import pandas


class Driver():
    
    def __init__(self,nb_traces=6):
              
        self.nb_traces = int(nb_traces)
        
        for i in range(1,self.nb_traces+1):
            setattr(self,f'trace{i}',Traces(self,i))
    
    
    ### User utilities
    def get_data_traces(self,traces=[],single=None):
        """Get all traces or the ones specified"""
        previous_trigger_state = self.get_previous_trigger_state()
        if single: self.single()
        self.ensure_scope_stopped()
        if traces == []: traces = list(range(1,self.nb_traces+1))
        for i in traces:
            getattr(self,f'trace{i}').get_data()
        self.set_previous_trigger_state(previous_trigger_state)
        
    def save_data_traces(self,filename,traces=[],FORCE=False):
        if traces == []: traces = list(range(1,self.nb_traces+1))
        for i in traces:
            getattr(self,f'trace{i}').save_data(filename=filename,FORCE=FORCE)
        
    ### Trigger functions
    def single(self):
        self.write('*TRG')
    def get_previous_trigger_state(self):
        return self.query('INIT:CONT?')
    def set_previous_trigger_state(self,previous_trigger_state):
        self.write(f'INIT:CONT {previous_trigger_state}')
    def ensure_scope_stopped(self):
        self.query('INIT:CONT OFF;*OPC?')
        
    def get_driver_model(self):
        model = []
        for i in range(1,self.nb_channels+1):
            model.append({'element':'module','name':f'line{i}','object':getattr(self,f'trace{i}'), 'help':'Traces'})
        model.append({'element':'action','name':'ensure_scope_stopped','do':self.ensure_scope_stopped,'help':'Returns only if the scope is stopped'})
        model.append({'element':'action','name':'single','do':self.single,'help':'Set single mode for trigger'})
        return model


#################################################################################
############################## Connections classes ##############################
class Driver_VXI11(Driver):
    def __init__(self, address='192.168.0.14', **kwargs):
        import vxi11 as v
    
        self.inst = v.Instrument(address)
        Driver.__init__(self, **kwargs)

    def write(self, msg):
        self.sock.write(msg)
    def read(self):
        return self.sock.read()
    def query(self,msg):
        """Sends question and returns answer"""
        self.write(msg)
        return(self.read())
    def close(self):
        self.inst.close()
############################## Connections classes ##############################
#################################################################################


class Traces():
    def __init__(self,dev,trace):
        self.trace     = int(trace)
        self.dev       = dev
        self.data_dict = {}
        
    def get_data(self):
        self.data        = eval(self.dev.query(f"TRAC:DATA? TRACE{self.trace}"))
        self.frequencies = self.get_frequencies(self.data)
        return self.frequencies,self.data
    def get_data_dataframe(self):
        frequencies,data              = self.get_data()
        self.data_dict['frequencies'] = self.frequencies
        self.data_dict['amplitude']   = self.data
        return pandas.DataFrame(self.data_dict)
    
    def get_start_frequency(self):
        return float(self.dev.query("SENS:FREQ:START?"))
    def get_stop_frequency(self):
        return float(self.dev.query("SENS:FREQ:STOP?"))
    def get_frequencies(self,data):
        start = self.get_start_frequency()
        stop  = self.get_stop_frequency()
        return linspace(start,stop,len(data))
    
    def save_data(self,filename,FORCE=False):
        temp_filename = f'{filename}_MXAN9020ATR{self.trace}.txt'
        if os.path.exists(os.path.join(os.getcwd(),temp_filename)) and not(FORCE):
            print('\nFile ', temp_filename, ' already exists, change filename or remove old file\n')
            return
        
        savetxt(temp_filename,(self.frequencies,self.data))


    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'spectrum','read':self.get_data_dataframe,'type':pandas.DataFrame,'help':'Current spectrum'})
        return model
    



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:21:49 2019

@author: qchat
"""
import time
import threading
from PyQt5 import QtCore

class MonitorManager : 
    
    def __init__(self,gui):
        
        self.gui = gui
        
        # Configure a new thread
        self.thread = MonitorThread(self.gui.variable,self.gui.queue)
        self.thread.errorSignal.connect(self.error)
        
        
        
    def error(self,error):
        
        """ This function is called when the errorSignal of the thread is raised.
        It update the pause button and displays the error in the GUI """
        
        self.gui.pauseButton.setText('Resume')
        self.gui.statusBar.showMessage(f'Error : {error} ',10000)
        
    
    
    def start(self):
        
        """ This function start the thread """
        
        self.thread.start()
        
        
        
    def setDelay(self,value):
        
        """ Set the delay in the thread """

        self.thread.delay = value
        
        
        
    def getDelay(self):
        
        """ Returns the current delay of the thread """
        
        return self.thread.delay
        
        
    
    def isPaused(self):
        
        """ This function returns whether the thread is paused or not """
        
        return self.thread.pauseFlag.is_set()
        

            
    def resume(self):
        
        """ This function resume the monitoring """
        
        self.thread.pauseFlag.clear()
    
    
    
    def pause(self):
        
        """ This function pause the monitoring """
        
        self.thread.pauseFlag.set()          
        
    
    
    def close(self):
        
        """ This function stops the thread and wait for its complete deletion """
        
        self.resume()
        self.thread.stopFlag.set()
        self.thread.wait()





class MonitorThread(QtCore.QThread):
    
    """ This thread class is dedicated to read the variable, and send its data to GUI through a queue """
    
    errorSignal = QtCore.pyqtSignal(object) 

    def __init__(self,variable,queue):
        
        QtCore.QThread.__init__(self)
        self.variable = variable
        self.queue = queue
        
        self.pauseFlag = threading.Event()
        self.stopFlag = threading.Event()
        
        self.delay = 0
        
        
    def run(self):
        
        t_ini = time.time()
        
        pauseLength = 0
        pauseStartedTime = None      
        
        while self.stopFlag.is_set() is False :
                
            # If the thread just resume, take into account the delay it has been paused
            if pauseStartedTime is not None :
                pauseLength += time.time()-pauseStartedTime
                pauseStartedTime = None
            
            # Time measure
            now = time.time() - t_ini - pauseLength
            
            try : 
                
                # Measure variable
                value = self.variable()
                
                # Check type
                value = float(value)
                
                # Send signal new data
                self.queue.put([now,value])
                
            except Exception as e :
                
                self.errorSignal.emit(e)
                self.pauseFlag.set()
            
            # If not the thread may be too fast
            time.sleep(self.delay)
                
            # pause
            while self.pauseFlag.is_set() :
                if pauseStartedTime is None :
                    pauseStartedTime = time.time()
                time.sleep(0.1)
                


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- Newport smc100
"""



class Driver():

    slot_naming = 'slot<NUM> = <MODULE_NAME>'
    
    def __init__(self,**kwargs):
        
        # Submodules loading
        self.slot_names = {}
        prefix = 'slot'
        for key in kwargs.keys():
            if key.startswith(prefix) and not '_name' in key :
                slot_num = key[len(prefix):]
                module_name = kwargs[key].strip()
                module_class = globals()[f'Module_{module_name}']
                if f'{key}_name' in kwargs.keys() : name = kwargs[f'{key}_name']
                else : name = f'{key}_{module_name}'
                setattr(self,name,module_class(self,slot_num))
                self.slot_names[slot_num] = name

    def get_driver_model(self):
        
        model = []
        for name in self.slot_names.values():
            model.append({'element':'module','name':name,'object':getattr(self,name)})
        return model
    
#################################################################################
############################## Connections classes ##############################
class Driver_VISA(Driver):
    def __init__(self, address='GPIB0::2::INSTR', **kwargs):
        import visa
        
        rm = visa.ResourceManager()
        self.controller = rm.open_resource(address)
        self.controller.baud_rate = 57600
        self.controller.flow_control = visa.constants.VI_ASRL_FLOW_XON_XOFF
        self.controller.read_termination = '\r\n'
        
        Driver.__init__(self, **kwargs)
    

    def query(self,command) :
        return self.controller.query(command)
        
    def write(self,command) :
        self.controller.write(command)

############################## Connections classes ##############################
#################################################################################


import time 

class Module_ILS100CC() :
    
    def __init__(self,dev,slot):
        
        self.dev = dev
        self.SLOT = str(slot)        
        
    def query(self,command,unwrap=True):
        result = self.dev.query(self.SLOT+command)
        if unwrap is True :
            try:
                prefix=self.SLOT+command[0:2]
                result = result.replace(prefix,'')
                result = result.strip()
                result = float(result)
            except:
                pass
            
        return result
    
    
    def write(self,command):
        self.dev.write(self.SLOT+command)
        
    def get_state(self):
        ans = self.query('TS?',unwrap=False)[-2:]
        if ans[0] == '0' :
            return 'NOTREF'
        elif ans == '14' :
            return 'CONF'
        elif ans in ['1E','1F'] :
            return 'HOMING'
        elif ans == '28' :
            return 'MOVING'
        elif ans[0] == '3' and ( ans[1].isalpha() is False)  :
            return 'READY'
        elif ans[0] == '3' and ( ans[1].isalpha() is True)  :
            return 'DISABLED'
        elif ans in ['46','47']:
            return 'JOGGING'
        else :
            return 'UNKNOWN'
        
    def set_ready_state(self):
        
        # On vérifie que l'on est pas dans le mode REF
        state=self.get_state()
        if state == 'JOGGING' : 
            self.write('JD') # Sortie du mode Jogging   
        elif state == 'DISABLED' : 
            self.write('MM1') # Sortie du mode disabled  
        self.wait_ready_state()

    def wait_ready_state(self):
        while self.get_state() != 'READY' :
            time.sleep(0.1)
            

    def get_id(self):
        return self.query('ID?')


    def get_position(self):
        return self.query('PA?')
    
    def set_position(self,value):
        self.set_ready_state()
        self.write('PA'+str(value))
        self.wait_ready_state()


    def set_acceleration(self,value):
        self.set_ready_state()
        self.write('AC'+str(value))


    def get_acceleration(self):
        return self.query('AC?')
    
    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'position','type':str,
                       'read':self.get_position,'write':self.set_position, 'help':'Initiates an absolute move.'})
        return model
            




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- 
"""

import os
import time
from numpy import frombuffer,int8,ndarray


class Driver():
    
    def __init__(self,nb_channels=4):
              
        self.nb_channels = int(nb_channels)
        self.type        = 'BYTE'
        
        self.write(':WAVeform:TYPE RAW')
        self.write(':WAVEFORM:BYTEORDER LSBFirst')
        self.write(':TIMEBASE:MODE MAIN')
        self.write(':WAV:SEGM:ALL ON')
        self.set_type(self.type)
        
        for i in range(1,self.nb_channels+1):
            setattr(self,f'channel{i}',Channel(self,i))
    
    
    ### User utilities
    def get_data_channels(self,channels=[]):
        """Get all channels or the ones specified"""
        self.stop()
        if channels == []: channels = list(range(1,self.nb_channels+1))
        for i in channels:
            getattr(self,f'channel{i}').get_data_raw()
            getattr(self,f'channel{i}').get_log_data()
        self.run()
        
    def save_data_channels(self,filename,channels=[],FORCE=False):
        if channels == []: channels = list(range(1,self.nb_channels+1))
        for i in channels:
            getattr(self,f'channel{i}').save_data_raw(filename=filename,FORCE=FORCE)
            getattr(self,f'channel{i}').save_log_data(filename=filename,FORCE=FORCE)
        
    ### Trigger functions
    def run(self):
        self.write(':RUN')
    def stop(self):
        self.write(':STOP')
    def single(self):
        self.write(':SINGLE')
        
    def set_type(self,val):
        """Argument type must be a string (BYTE or ASCII)"""
        self.type = val
        self.write(f':WAVEFORM:FORMAT {self.type}')
    def get_type(self):
        return self.type


    def get_driver_model(self):
        model = []
        for i in range(1,self.nb_channels+1):
            model.append({'element':'module','name':f'channel{i}','object':getattr(self,f'channel{i}'), 'help':'Channels'})
        model.append({'element':'variable','name':'encoding','write':self.set_type,'read':self.get_type, 'type':str,'help':'Set the data encoding too use. Accepted values are: BYTE, ASCII. Default value is BYTE'})
        model.append({'element':'action','name':'stop','do':self.stop,'help':'Set stop mode for trigger'})
        model.append({'element':'action','name':'run','do':self.run,'help':'Set run mode for trigger'})
        model.append({'element':'action','name':'single','do':self.single,'help':'Set single mode for trigger'})
        return model

#################################################################################
############################## Connections classes ##############################
class Driver_VXI11(Driver):
    def __init__(self, address='192.168.0.1', **kwargs):
        import vxi11 as v
    
        self.inst = v.Instrument(address)
        Driver.__init__(self, **kwargs)

    def read_raw(self,length=100000000):
        rep = self.inst.read_raw(length)
        return rep
    def read(self,length=100000000):
        rep = self.inst.read(length)
        return rep
    def write(self,cmd):
        self.inst.write(cmd)
    def close(self):
        self.inst.close()
############################## Connections classes ##############################
#################################################################################

class Channel():
    def __init__(self,dev,channel):
        self.channel = int(channel)
        self.dev  = dev
    
        
    def get_data_raw(self):
        self.dev.write(f':WAVEFORM:SOURCE CHAN{self.channel}')
        self.dev.write(':WAV:DATA?')
        self.data_raw = self.dev.read_raw()
        if typ=='BYTE':
            self.data_raw = self.data_raw[8:]
        self.data_raw = self.data_raw[:-1]
        return self.data_raw
    def get_log_data(self):
        self.dev.write(f':WAVEFORM:SOURCE CHAN{self.channel}')
        self.dev.write(f':WAVEFORM:PREAMBLE?')
        self.log_data = self.dev.read()
        return self.log_data
    def get_data(self):
        return frombuffer(self.get_data_raw(),int8)
        
    def save_data_raw(self,filename,FORCE=False):
        temp_filename = f'{filename}_DSO54853ACH{self.channel}'
        if os.path.exists(os.path.join(os.getcwd(),temp_filename)) and not(FORCE):
            print('\nFile ', temp_filename, ' already exists, change filename or remove old file\n')
            return
        f = open(temp_filename,'wb')# Save data
        f.write(self.data_raw)
        f.close()
    def save_log_data(self,filename,FORCE=False):
        temp_filename = f'{filename}_DSO54853ACH{self.channel}.log'
        if os.path.exists(os.path.join(os.getcwd(),temp_filename)) and not(FORCE):
            print('\nFile ', temp_filename, ' already exists, change filename or remove old file\n')
            return
        f = open(temp_filename,'w')
        f.write(self.log_data)
        f.close()
        
    def get_data_numerical(self):
        return array_of_float
    def save_data_numerical(self):
        return array_of_float


    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'trace_raw','type':bytes,'read':self.get_data_raw,'help':'Get the current trace in bytes'})
        model.append({'element':'variable','name':'trace','type':ndarray,'read':self.get_data,'help':'Get the current trace in a numpy array of integers'})
        return model



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 17 19:24:07 2019

@author: qchat
"""

from threading import Thread
import autolab

def startup() :
    
    ''' Send a 'startup' anonymous event to google analytics '''

    send('startup')
    
    
    
def send(action):
    
    ''' Send a <action> anonymous event to google analytics '''
    
    if is_stats_enabled():
        StatisticsThread(action).start()
        
        
        
def set_stats_enabled(state):
    
    """ This function enable or disable the anonymous statistics feature of Autolab """
    
    assert isinstance(state,bool), "The state has to be a boolean."
    autolab._config.set_value('stats','enabled',str(int(state)))
    
    
    
def is_stats_enabled():
    
    """ This function the activation state of the anonymous statistics feature of Autolab """
    
    return autolab._config.get_value('stats','enabled') == '1' 


def get_explanation():
    
    return '''At startup, Autolab is configured to send only once a completely anonymous signal (sha256 hashed ID) over internet for statistics of use. This helps the authors to have a better understanding of how the package is used worldwide. No personal data is transmitted during this process. Also, this is done in background, with no impact on the performance of Autolab.'''
class StatisticsThread(Thread) :
    
    def __init__(self,action) :
        Thread.__init__(self)
        self.action = action
        
    def run(self):
        
        ''' Send an anonymous event to google analytics '''
        
        import requests
        import hashlib
        import uuid
        import socket
        
        # Fabrication of a unique anonymous identifier based on the hostname and mac address
        uid = socket.gethostname()+hex(uuid.getnode())
        uid = hashlib.sha256(uid.encode()).hexdigest()   

        # Preparing payload
        data = {
            'v': '1',  # API Version.
            't': 'event',  # Event hit type.
            'tid': 'UA-152731770-3',  # Tracking ID / Property ID of the Google Analytics property
            'cid': uid, # Unique Anonymous Client Identifier
            'ec': 'package',  # Event category.
            'ea': self.action,  # Event action.
            'el': autolab.__version__ # Event label
        }
        
        try : 
            requests.post('https://www.google-analytics.com/collect', data=data)
        except : pass
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- 
"""


class Driver():
    
    def __init__(self):
        
        self.conv = ['T0','T1','A','B','C','D','E','F','G','H']
        self.conv2 = ['T0','AB','CD','EF','GH']
        self.delay_channels = ''
        
        for i in self.conv2:
            setattr(self,f'output{i}',Output(self,i))
    
        
    def set_frequency(self,frequency):
        self.write(f'TRAT{frequency}')
    def get_frequency(self):
        return float(self.query('TRAT?'))
    
    def set_delay_channels(self,channels):
        assert len(channels)==1 or len(channels)==2, f"Argument channels must be a str of length either 1 or 2 in {self.conv[2:]}"
        self.delay_channels = channels
        
    def set_delay(self, delay,channels=None):
        if channels is None: channels = self.delay_channels
        assert len(channels)==1 or len(channels)==2, "You may want to set_delay_channels first"
        if len(channels) == 2:
            ch1 = str(self.conv.index(channels[0]))
            ch2 = str(self.conv.index(channels[1]))
        elif len(channels) == 1:
            ch1 = '0'
            ch2 = str(self.conv.index(channels))
        self.write(f'DLAY{ch2},{ch1},{delay}')

    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'frequency','type':float,'unit':'Hz','write':self.set_frequency,'read':self.get_frequency,'help':'Output frequency.'})
        return model
        
#################################################################################
############################## Connections classes ##############################
class Driver_VXI11(Driver):
    def __init__(self,address='192.168.0.4', **kwargs):
        import vxi11 as v

        self.inst = v.Instrument(address)
        Driver.__init__(self)

    def query(self, cmd, nbytes=1000000):
        """Send command 'cmd' and read 'nbytes' bytes as answer."""
        self.write(cmd+'\n')
        r = self.read(nbytes)
        return r
    def read(self,nbytes=1000000):
        return self.inst.read(nbytes)
    def write(self,cmd):
        self.inst.write(cmd)
    def close(self):
        self.inst.close()
############################## Connections classes ##############################
#################################################################################


class Output():
    def __init__(self,dev,output):
        self.output = str(output)
        self.dev     = dev
    
    def set_amplitude(self,amplitude):
        self.dev.write(f'LAMP{self.dev.conv2.index(self.output)},{amplitude}')
    def get_amplitude(self):
        return float(self.dev.query(f'LAMP?{self.dev.conv2.index(self.output)}'))
    def set_polarity(self,polarity):
        self.dev.write(f'LPOL{self.dev.conv2.index(self.output)},{polarity:g}')
    def get_polarity(self):
        return bool(int(self.dev.query(f'LPOL?{self.dev.conv2.index(self.output)}')))
    def set_offset(self,offset):
        self.dev.write(f'LOFF{self.dev.conv2.index(self.output)},{offset}')
    def get_offset(self):
        return float(self.dev.query(f'LOFF?{self.dev.conv2.index(self.output)}'))
    
    def set_delay(self,delay):
        self.dev.set_delay(delay,channels=self.output)
            
    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'amplitude','type':float,'unit':'V','write':self.set_amplitude,'read':self.get_amplitude,'help':'Voltage amplitude.'})
        model.append({'element':'variable','name':'polarity','type':bool,'read':self.get_polarity,'write':self.set_polarity,'help':'Polarity state.'})
        model.append({'element':'variable','name':'offset','type':float,'unit':'V','write':self.set_offset,'read':self.get_offset,'help':'Voltage offset'})
        model.append({'element':'variable','name':'delay','type':float,'unit':'s','write':self.set_delay,'help':'Set the delay between the two channels of the given output'})
        return model
    



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 13 10:25:49 2019

@author: quentin.chateiller
"""


import autolab
from .elements import Module
from .utilities import emphasize

# Storage of the devices
DEVICES = {}


def get_element_by_address(address):
    
    """ Returns the Element located at the provided address """
    
    address = address.split('.')
    
    try :
        element = get_device(address[0])
        for addressPart in address[1:] :
            element = getattr(element,addressPart)
        return element
    except :
        pass
        


def get_device(name, **kwargs):
    
    ''' Returns the Device associated to device_name. Load it if not already done.'''
    
    if name in DEVICES.keys() :
        assert len(kwargs)==0, f'You cannot change the configuration of an existing Device. Close it first & retry, or remove the provided configuration.'
    else : 
        instance = autolab.get_driver(name,**kwargs)
        DEVICES[name] = Device(name,instance)

    return DEVICES[name]
    
    
    
    
# =============================================================================
# DEVICES LIST HELP
# =============================================================================
        


def list_devices():
    
    ''' Returns the list of the loaded devices '''
    
    return list(DEVICES.keys())





# =============================================================================
# DEVICE CLASS
# =============================================================================

class Device(Module):
    
    def __init__(self,device_name,instance):
                
        Module.__init__(self,None,{'name':device_name,
                                   'object':instance,
                                   'help':f'Device {device_name}'})
        
    def close(self):
        
        """ This function close the connection of the current physical device """
        
        try : self.instance.close()
        except : pass
        del DEVICES[self.name]
        
        
    def __dir__(self):
        
        """ For auto-completion """
        
        return  self.list_modules() + self.list_variables() + self.list_actions() + ['close','help','instance']
 
        
    
    
    





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

category = 'Optical source'

class Driver_parser():
    def __init__(self, Instance, name, **kwargs):
        self.name     = name
        self.Instance = Instance
        
        
    def add_parser_usage(self,message):
        """Usage to be used by the parser"""
        usage = f"""
{message}

----------------  Examples:  ----------------

usage:    autolab driver [options] args
            
    autolab driver -D {self.name} -C VISA -A GPIB0::2::INSTR -w 1550.2 -p 10
    load {self.name} driver using VISA communication protocol with address GPIB... and set the wavelength to 1550.2nm and the power to 10mW.
    
    autolab driver -D nickname -a 100
    Use now the device nickname as defined in local_config.ini and set the pump current to 100mA.

    Note: you should provide either current or power at a time (same thing applies to frequency and wavelength)
            """
        return usage
    
    def add_parser_arguments(self,parser):
        """Add arguments to the parser passed as input"""
        parser.add_argument("-a", "--current", type=str, dest="current", default=None, help="Set the pump current in mA." )
        parser.add_argument("-p", "--power", type=str, dest="power", default=None, help="Set the output power in mW." )
        parser.add_argument("-f", "--frequency", type=str, dest="frequency", default=None, help="Set the operating frequency in GHz." )
        parser.add_argument("-w", "--wavelength", type=str, dest="wavelength", default=None, help="Set the operating wavelength in nm." )
        return parser

    def do_something(self,args):
        if args.wavelength: 
            assert args.wavelength or args.frequency, "Please provide EITHER wavelength OR frequency"
            getattr(self.Instance,'set_wavelength')(args.wavelength)
        elif args.frequency: 
            getattr(self.Instance,'set_frequency')(args.frequency)
        if args.current: 
            assert args.current or args.power, "Please provide EITHER current OR power"
            getattr(self.Instance,'set_intensity')(args.current)
        elif args.power: getattr(self.Instance,'set_power')(args.power)
        

    def exit(self):
        self.Instance.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- 
"""

import os
from numpy import savetxt,linspace
import pandas
import time


class Driver():
    
    def __init__(self):
        
        for i in ['A','B','C']:
            setattr(self,f'trace{i}',Traces(self,i))
    
    
    ### User utilities
    def get_data_traces(self,traces=[],single=None):
        """Get all traces or the ones specified"""
        self.stop()
        if single: self.single()   # must verify whether finished sweeping
        while not self.is_scope_stopped(): time.sleep(0.05)
        if traces == []: traces = ['A','B','C']
        for i in traces:
            getattr(self,f'trace{i}').get_data()
        self.run()
        
    def save_data_traces(self,filename,traces=[],FORCE=False):
        if traces == []: traces = ['A','B','C']
        for i in traces:
            getattr(self,f'trace{i}').save_data(filename=filename,FORCE=FORCE)
        
    ### Trigger functions
    def single(self):
        s = self.write("SGL")
        return s
    def run(self):
        self.write('RPT')
    def stop(self):
        self.write('STP')
    def is_scope_stopped(self):
        return '0' in self.query('SWEEP?')
    
    ### Amplitude functions
    def set_ref_level(self,val):
        self.write(f'REFL{val}')
    def get_ref_level(self):
        return float(self.query(f'REFL?'))
    def set_level_scale(self,val):
        self.write(f'LSCL{val}')
    def get_level_scale(self):
        return float(self.query(f'LSCL?'))
        
    def get_driver_model(self):
        model = []
        for i in ['A','B','C']:
            model.append({'element':'module','name':f'line{i}','object':getattr(self,f'trace{i}'), 'help':'Traces'})
        model.append({'element':'variable','name':'reference_level','read':self.get_ref_level,'write':self.set_ref_level, 'type':float,'help':'Reference level'})
        model.append({'element':'variable','name':'level_scale','read':self.get_level_scale,'write':self.set_level_scale, 'type':float,'help':'Level scale'})
        model.append({'element':'variable','name':'is_stopped','read':self.is_scope_stopped, 'type':bool,'help':'Query whether scope is stopped'})
        model.append({'element':'action','name':'run','do':self.run,'help':'Set run mode for trigger'})
        model.append({'element':'action','name':'single','do':self.single,'help':'Set single mode for trigger'})
        model.append({'element':'action','name':'stop','do':self.stop,'help':'Set stop mode for trigger'})
        return model
        

#################################################################################
############################## Connections classes ##############################
class Driver_VISA(Driver):
    def __init__(self, address='GPIB0::2::INSTR', **kwargs):
        import visa as v
    
        r          = v.ResourceManager()
        self.scope = r.get_instrument(address)
        Driver.__init__(self)

    def query(self,query,length=1000000):
        self.write(query)
        r = self.read(length=length)
        return r
    def write(self,query):
        self.string = query + '\n'
        self.scope.write(self.string)
    def read(self,length=10000000):
        rep = self.scope.read()
        return rep
    def close(self):
        try : self.scope.close()
        except : pass
    
class Driver_GPIB(Driver):
    def __init__(self,address=19,board_index=0,**kwargs):
        import Gpib
        
        self.inst = Gpib.Gpib(int(board_index),int(address))
        Driver.__init__(self)
    
    def query(self,query):
        self.write(query)
        return self.read()
    def write(self,query):
        self.inst.write(query)
    def read(self,length=1000000000):
        msg = self.inst.read(length)
        return ''.join([chr(lett) for lett in msg]).strip('\r\n')  # from binary to string (used instead of decode() beacause of some weird bytes that may be there)
    def close(self):
        """WARNING: GPIB closing is automatic at sys.exit() doing it twice results in a gpib error"""
        #Gpib.gpib.close(self.inst.id)
        pass
############################## Connections classes ##############################
#################################################################################


class Traces():
    def __init__(self,dev,trace):
        self.trace     = str(trace)
        self.dev       = dev
        self.data_dict = {}
        
    def get_data(self):
        self.data        = self.dev.query(f"LDAT{self.trace}").split(',')[1:]
        self.data        = [float(val.replace(' ','')) for val in self.data]
        self.frequencies = self.get_frequencies(self.data)
        return self.frequencies,self.data
    
    def get_data_dataframe(self):
        frequencies,data              = self.get_data()
        self.data_dict['frequencies'] = self.frequencies
        self.data_dict['amplitude']   = self.data
        return pandas.DataFrame(self.data_dict)
    
    def save_data(self,filename,FORCE=False):
        temp_filename = f'{filename}_AQ6315ATR{self.trace}.txt'
        if os.path.exists(os.path.join(os.getcwd(),temp_filename)) and not(FORCE):
            print('\nFile ', temp_filename, ' already exists, change filename or remove old file\n')
            return
        
        savetxt(temp_filename,(self.frequencies,self.data))

    def set_start_wavelength(self,value):
        self.dev.write(f'STAWL {value}')
    def set_stop_wavelength(self,value):
        self.dev.write(f'STPWL {value}')
    def get_start_frequency(self):
        return float(self.dev.query("STAWL?"))
    def get_stop_frequency(self):
        return float(self.dev.query("STPWL?"))
    def get_frequencies(self,data):
        start = self.get_start_frequency()
        stop  = self.get_stop_frequency()
        return linspace(start,stop,len(data))
    
    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'start_wavelength','write':self.set_start_wavelength,'type':float,'help':'Start wavelength of the window'})
        model.append({'element':'variable','name':'stop_wavelength','write':self.set_stop_wavelength,'type':float,'help':'Stop wavelength of the window'})
        model.append({'element':'variable','name':'start_frequency','read':self.get_start_frequency,'type':float,'help':'Start frequency of the window'})
        model.append({'element':'variable','name':'stop_frequency','read':self.get_stop_frequency,'type':float,'help':'Stop frequency of the window'})
        model.append({'element':'variable','name':'spectrum','read':self.get_data_dataframe,'type':pandas.DataFrame,'help':'Current spectrum'})
        return model
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- 
"""

import os
import time
from numpy import savetxt,ndarray


class Driver():
    
    def __init__(self,camera='Camera1'):
        self.camera = camera
        
    def set_camera(self,camera):
        self.write(f'CONNECT={camera}')
        self.camera = camera
    def set_nb_frames(self,nb_frames):
        self.write(f'NBFRAMES={nb_frames}')
    def set_exposure(self,exposure):
        self.write(f'EXPTIME={exposure}')
        
    def get_data(self):
        self.data = eval(self.query('DATA?'))
    def get_exposure(self):
        return float(self.query('EXPTIME?'))
    def get_nb_frames(self):
        return int(self.query('NBFRAMES?'))
    def list_cameras(self):
        return self.query('LISTCAMS?')
    
    def enable_auto_exposure(self):
        self.write('AUTOEXPEN')
    def disable_auto_exposure(self):
        self.write('AUTOEXPDIS')
    def save_data_remote(self):
        self.write('SAVEDATA')
    def save_data(self,filename,FORCE=False):
        temp_filename = f'{filename}_SPECTRO32{self.camera}.txt'
        if os.path.exists(os.path.join(os.getcwd(),temp_filename)) and not(FORCE):
            print('\nFile ', temp_filename, ' already exists, change filename or remove old file\n')
            return
        self.lambd=[list(range(len(self.data[0])))]
        assert self.lambd, 'You may want to get_data before saving ...'
        [self.lambd.append(self.data[i]) for i in range(len(self.data))]
        f = savetxt(temp_filename,self.lambd)  ## squeeze
        print('Spectro32 data saved')


    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'exposure','type':float,'read':self.get_exposure,'write':self.set_exposure,'help':'Manage the camera exposure time'})
        model.append({'element':'variable','name':'nb_frames','type':int,'read':self.get_nb_frames,'write':self.set_nb_frames,'help':'Manage the number of frames to be acquired'})
        model.append({'element':'variable','name':'camera','type':str,'read':self.list_cameras,'write':self.set_camera,'help':'Manage cameras'})
        model.append({'element':'variable','name':'trace','type':ndarray,'read':self.get_data,'help':'Get the current trace in a numpy array'})
        model.append({'element':'action','name':'enable_auto_exposure','do':self.enable_auto_exposure,'help':'Enable auto exposure mode'})
        model.append({'element':'action','name':'disable_auto_exposure','do':self.disable_auto_exposure,'help':'Disable auto exposure mode'})
        return model


#################################################################################
############################## Connections classes ##############################
class Driver_SOCKET(Driver):
    def __init__(self,address='192.168.0.2',camera='Camera1',**kwargs):
        import socket 
        
        self.PORT        = 5005
        self.BUFFER_SIZE = 10000
        
        self.controller = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.controller.connect((address,int(self.PORT)))
        
        Driver.__init__(self,camera,**kwargs)
        
    def write(self,command):
        self.controller.send(command.encode())
        self.recv_end(self.controller)
    def query(self,command):
        self.controller.send(command.encode())
        data = self.recv_end(self.controller)
        return data
    def recv_end(self,the_socket,end='\n'):
        total_data=[];data=''
        while True:
                data=the_socket.recv(self.BUFFER_SIZE).decode()
                if end in data:
                    total_data.append(data[:data.find(end)])
                    break
                total_data.append(data)
                if len(total_data)>1:
                    #check if end_of_data was split
                    last_pair=total_data[-2]+total_data[-1]
                    if end in last_pair:
                        total_data[-2]=last_pair[:last_pair.find(end)]
                        total_data.pop()
                        break
        return ''.join(total_data)
    def close(self):
        try :
            self.controller.close()
        except :
            pass
        self.controller = None

class Driver_LOCAL(Driver):
    def __init__(self,camera='Camera1',**kwargs):

        from spectro32_utilities.spectro32_driver import Driver
        self.controller = Driver()

        Driver.__init__(self,camera,**kwargs)
              
    def write(self,command):
        self.controller.command(command)
    def query(self,command):
        return self.controller.command(command)
    def close(self):
        self.controller = None
############################## Connections classes ##############################
#################################################################################





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Supported instruments (identified):
- 
"""


class Driver() :
    
    def __init__(self):
        pass

    def get_id(self):
        return self.query('*IDN?')
    
    def set_resol(self,value):
        self.write(f'RES {value}')
        self.query('*OPC?')
    
    def get_resol(self):
        return self.query('RES?')
    
    def set_span(self, value):
        self.write(f'SPN {value}')
        self.query('*OPC?')
        
    def get_span(self):
        return self.query('SPN?')
    
    def single_sweep(self):    #single sweep measurement
        self.write('SSI')
        self.query('*OPC?')
        
    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'resolution','write':self.set_resol,'read':self.get_resol,'type':float,'help':"Resolution"})
        model.append({'element':'variable','name':'span','write':self.set_span,'read':self.get_span,'type':float,'help':"Span"})
        model.append({'element':'action','name':'single_sweep','do':self.single_sweep,'help':"Perform a single sweep"})

        return model
           
#################################################################################
############################## Connections classes ##############################
class Driver_VISA(Driver):
    def __init__(self,address='GPIB0::1::INSTR', **kwargs):
        import visa
        
        rm = visa.ResourceManager()
        self.controller = rm.open_resource(address)
    
        Driver.__init__(self)
    
    def close(self):
        self.controller.close()
    def query(self,command):
        return self.controller.query(command)
    def write(self,command):
        self.controller.write(command)
############################## Connections classes ##############################
#################################################################################
            


        
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  3 20:06:08 2019

@author: quentin.chateiller
"""

import time


class Driver():

    def __init__(self):
    
        # Initialisation
        self.write('*CLS')      # Clear status registers
        self.write('DELAY 0')   # User commands not delayed
        
        # Subdevices
        self.tec = TEC(self)
        self.las = LAS(self)
        
        
    def get_id(self):
        return self.query('*IDN?')
    
    
    def get_driver_model(self):
        model = []
        model.append({'element':'module','name':'tec','object':self.tec})
        model.append({'element':'module','name':'las','object':self.las})
        return model
    
    
#################################################################################
############################## Connections classes ##############################
class Driver_VISA(Driver):
    def __init__(self, address='GPIB0::2::INSTR',**kwargs):
        import visa
        
        self.TIMEOUT = 15000 #ms
        
        # Instantiation
        rm = visa.ResourceManager()
        self.controller = rm.get_instrument(address)
        self.controller.timeout = self.TIMEOUT
        self.controller.read_termination='\n'
        self.controller.write_termination='\n'
        self.controller.baud_rate = 19200
        #self.controller.query_delay = 0.05
        
        Driver.__init__(self)

    def close(self):
        try : self.controller.close()
        except : pass

    def query(self,command):
        return self.controller.query(command)
    
    def write(self,command):
        self.controller.write(command)
        time.sleep(0.01)
        
class Driver_GPIB(Driver):
    def __init__(self,address=2,board_index=0,**kwargs):
        import Gpib
        
        self.inst = Gpib.Gpib(int(board_index),int(address))
        Driver.__init__(self)
    
    def query(self,query):
        self.write(query)
        return self.read()
    def write(self,query):
        self.inst.write(query)
    def read(self,length=1000000000):
        return self.inst.read().decode().strip('\r\n')
    def close(self):
        """WARNING: GPIB closing is automatic at sys.exit() doing it twice results in a gpib error"""
        #Gpib.gpib.close(self.inst.id)
        pass
############################## Connections classes ##############################
#################################################################################


class LAS():
    
    category = 'Electrical source'
    
    def __init__(self,dev):
        
        self.dev = dev
        self.PRECISION = 0.1

    def write(self,command):
        self.dev.write(command)
        
    def query(self,command):
        return self.dev.query(command)
    
    
    def get_current(self):
        return float(self.query('LAS:LDI?'))

    def get_current_setpoint(self):
        return float(self.query('LAS:SET:LDI?'))
        
    def set_current(self,value):
        assert isinstance(float(value),float)
        value = float(value)
        self.write(f'LAS:LDI {value}')
        #self.query('*OPC?')
        if value > 0 :
            if self.is_enabled() is False :
                self.set_enabled(True)
            self.set_mode('ILBW')
            self.wait_for_convergence(self.get_current,value)
        elif value == 0 :
            if self.is_enabled() is True :
                self.set_enabled(False)




    def set_power(self,value):
        assert isinstance(float(value),float)
        value = float(value)
        self.write(f'LAS:MDP {value}')
        #self.query('*OPC?')
        if value > 0 :
            if self.is_enabled() is False :
                self.set_enabled(True)
            self.set_mode('MDP')
            self.wait_for_convergence(self.get_power,value)
        elif value == 0 :
            if self.is_enabled() is True :
                self.set_enabled(False)

    def get_power(self):
        #self.query('*OPC?')
        return float(self.query('LAS:MDP?'))

    def get_power_setpoint(self):
        return float(self.query('LAS:SET:MDP?'))
    
    
    
    
    def set_enabled(self,value):
        assert isinstance(value,bool)
        self.write(f'LAS:OUT {int(value)}')
        #self.query('*OPC?')
        if value is True :
            mode = self.query('LAS:MODE?')
            if mode.startswith('I'):
                self.wait_for_convergence(self.get_current,
                                        self.get_current_setpoint())
            elif mode.startswith('MD'):
                self.wait_for_convergence(self.get_power,
                                        self.get_power_setpoint())
        
    def is_enabled(self):
        return bool(int(self.query('LAS:OUT?')))

    
    
    
    def wait_for_convergence(self,func,setpoint):
        tini = time.time()
        while True :
            try : 
                if abs(func()-setpoint) < self.PRECISION*setpoint :
                    break
                else :
                    time.sleep(0.5)
            except :
                pass
            if time.time() - tini > 5 :
                break
            
            
            
            
    def set_mode(self,mode):
        assert isinstance(mode,str)
        curr_mode = self.get_mode()
        enabled_mode = self.is_enabled()
        if curr_mode != mode :
            self.write(f'LAS:MODE:{mode}')
            #self.query('*OPC?')
            self.set_enabled(enabled_mode)
            
    def get_mode(self):
        return self.query('LAS:MODE?')
        
        
    
    
    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'current_setpoint','type':float,'unit':'mA','read':self.get_current_setpoint,'help':'Current setpoint'})
        model.append({'element':'variable','name':'current','type':float,'unit':'mA','read':self.get_current,'write':self.set_current,'help':'Current'})
        model.append({'element':'variable','name':'power_setpoint','type':float,'unit':'mW','read':self.get_power_setpoint,'help':'Output power setpoint'})
        model.append({'element':'variable','name':'power','type':float,'unit':'mW','write':self.set_power,'read':self.get_power,'help':'Output power'})
        model.append({'element':'variable','name':'output','type':bool,'read':self.is_enabled,'write':self.set_enabled,'help':'Output state'})
        model.append({'element':'variable','name':'mode','type':str,'read':self.get_mode,'write':self.set_mode,'help':'Control mode'})
        return model
    
    



class TEC():
    
    category = 'Temperature controller'
    
    def __init__(self,dev):
        
        self.dev = dev
        self.PRECISION = 0.05



    def write(self,command):
        self.dev.write(command)
        
    def query(self,command):
        return self.dev.query(command)
    
    
    

    def get_resistance(self):
        return float(self.query('TEC:R?'))



    def set_gain(self,value):
        assert isinstance(int(value),int)
        value = int(value)
        self.write(f'TEC:GAIN {value}')
        #self.query('*OPC?')
        
    def get_gain(self):
        return int(float(self.query('TEC:GAIN?')))
   
    

    
    def get_current(self):
        return float(self.query('TEC:ITE?'))

    def get_current_setpoint(self):
        return float(self.query('TEC:SET:ITE?'))
        
    def set_current(self,value):
        assert isinstance(float(value),float)
        value = float(value)
        self.write(f'TEC:ITE {value}')
        #self.query('*OPC?')
        if value > 0 :
            if self.is_enabled() is False :
                self.set_enabled(True)
            self.set_mode('ITE')
            self.wait_for_convergence(self.get_current,value)
        elif value == 0 :
            if self.is_enabled() is True :
                self.set_enabled(False)




    def set_temperature(self,value):
        assert isinstance(float(value),float)
        value = float(value)
        self.write(f'TEC:T {value}')
        #self.query('*OPC?')
        if value > 0 :
            if self.is_enabled() is False :
                self.set_enabled(True)
            self.set_mode('T')
            self.wait_for_convergence(self.get_temperature,value)
        elif value == 0 :
            if self.is_enabled() is True :
                self.set_enabled(False)

    def get_temperature(self):
        return float(self.query('TEC:T?'))

    def get_temperature_setpoint(self):
        return float(self.query('TEC:SET:T?'))
    
    

    
    
    def set_enabled(self,value):
        assert isinstance(value,bool)
        self.write(f'TEC:OUT {int(value)}')
        #self.query('*OPC?')
        if value is True :
            mode = self.query('TEC:MODE?')
            if mode.startswith('I'):
                self.wait_for_convergence(self.get_current,
                                        self.get_current_setpoint())
            elif mode.startswith('T'):
                self.wait_for_convergence(self.get_temperature,
                                        self.get_temperature_setpoint())
        
    def is_enabled(self):
        return bool(int(self.query('TEC:OUT?')))
    
    
    
    
    
    def wait_for_convergence(self,func,setpoint):
        tini = time.time()
        while True :
            try : 
                if abs(func()-setpoint) < self.PRECISION*setpoint :
                    break
                else :
                    time.sleep(0.5)
            except :
                pass
            if time.time() - tini > 5 :
                break
            
            
            
            
    def set_mode(self,mode):
        assert isinstance(mode,str)
        curr_mode = self.get_mode()
        enabled_mode = self.is_enabled()
        if curr_mode != mode :
            self.write(f'TEC:MODE:{mode}')
            #self.query('*OPC?')
            self.set_enabled(enabled_mode)
            
    def get_mode(self):
        return self.query('TEC:MODE?')
    
    
    
    def get_driver_model(self):
        model = []
        model.append({'element':'variable','name':'resistance','type':float,'read':self.get_resistance,'help':'Resistance'})
        model.append({'element':'variable','name':'gain','type':int,'read':self.get_gain,'write':self.set_gain,'help':'Gain'})
        model.append({'element':'variable','name':'current_setpoint','type':float,'unit':'mA','read':self.get_current_setpoint,'help':'Current setpoint'})
        model.append({'element':'variable','name':'current','type':float,'unit':'mA','read':self.get_current,'write':self.set_current,'help':'Current'})
        model.append({'element':'variable','name':'temperature_setpoint','type':float,'unit':'°C','read':self.get_temperature_setpoint,'help':'Temperature setpoint'})
        model.append({'element':'variable','name':'temperature','type':float,'unit':'°C','read':self.get_temperature,'write':self.set_temperature,'help':'Actual temperature'})
        model.append({'element':'variable','name':'output','type':bool,'read':self.is_enabled,'write':self.set_enabled,'help':'Output state'})
        model.append({'element':'variable','name':'mode','type':str,'read':self.get_mode,'write':self.set_mode,'help':'Control mode'})
        return model
    
    

        
